import bge
import json

#import profile

_HAT_DICT = {0:None,
            1:'Up',
            2:'Right',
            4:'Down',
            8:'Left',
            3:'UpRight',
            6:'DownRight',
            12:'DownLeft',
            9:'UpLeft'}

_key_map = {}
_action_dict = {}
_callback_dict = {}

center_mouse = False

#def update():
#    profile.run('import bgeinput; bgeinput.update2()')

def update():
    '''Should be run every frame to update the internal state'''
    a = _get_axis_input()
    b = _get_button_input()
    
    for action in _key_map:
        _action_dict[action] = sum([a[action], b[action]])
        if action in _callback_dict:
            funct = _callback_dict[action][0]
            if funct != None:
                args = _callback_dict[action][1]
                funct(action, _action_dict[action], *args)

def load_input_map(f):
    '''Loads a keymap from file f'''
    global _key_map
    _key_map = json.load(open(f))

def save_input_map(f):
    '''Saves a keymap to file f
    Note: Erases contents of file'''
    open(f, 'w').write(json.dumps(_key_map))    
    
def set_callback(action, function=None, args=[]):
    '''Sets a function to be called upon a particular action. 
    To clear a callback use:
    set_callback(action, None)'''
    if function == None:
        del _callback_dict[action]
    else:
        _callback_dict[action] = [function, args]
    
def get_callback(action):
    '''Returns the callback and arguments for the specified callback'''
    return _callback_dict[action]

def add_mapping(action, key, sensitivity=0.5):
    '''Binds key to action.
    Inputs:
        - action - A string representing the desired action
        - key - A string representing the key (eg WKEY)
        - sensitivity (optional) - The sensitivity of an axis or ramping on a 
            button
        - direction (optional) - Invert an axis or the direction of a keypress 
            (ie so a button can have negative motion)'''
    _key_map[action] = _key_map.get(action, dict())
    _key_map[action][key] = sensitivity

def remove_mapping(action, key):
    '''Removes key from action'''
    del _key_map[action][key]

def get_mappings(action):
    '''Returns all mappings associated with the specified action'''
    return _key_map[action]

def state(action):
    '''Returns the state of action as defined in the keymap as a float.
    Returns None if the action doesn't exist.'''
    return _action_dict.get(action, None)

def get_active_inputs():
    '''Returns any active buttons, axis etc. Useful for GUI's to a keymapper
    Can be used like:
        bgeinput.add_mapping('Move', bgeinput.get_active_inputs()[0])'''
    active_buttons =  _get_key_buttons() + _get_mouse_buttons() + _get_joystick_buttons()
    active_axis = []
    all_axis = {}
    all_axis.update(_get_mouse_axis())
    all_axis.update(_get_joy_axis())
    
    for axis in all_axis:
        if all_axis[axis] > 0.2:
            active_axis.append(axis)
    return active_buttons + active_axis

def set_center_mouse(state):
    '''Mouselook requires the mouse to be centered. This function allows mouse 
    centering to be turned on and off (ie for menus)'''
    global center_mouse
    center_mouse = state

def get_mouse_over(scene=None):
    '''Returns the object the mouse is currently over'''
    if scene == None:
        scene = bge.logic.getCurrentScene()
    cam = scene.active_camera
    x,y = bge.logic.mouse.position
    return cam.getScreenRay(x,y,cam.far)

def get_typed_characters():
    '''Returns any typing character the user has typed this frame.'''
    active = get_active_inputs()
    if 'LEFTSHIFTKEY' in active or 'RIGHTSHIFTKEY' in active:
        shift = 1
    else:
        shift = 0
    
    string = ''
    for i in active:
        string += bge.events.EventToCharacter(bge.events.__dict__[i], shift)
    return string
    #Should rewrite using get_active_inputs
    '''
    pressed = list(bge.logic.keyboard.active_events.keys())
    key_dict = {bge.events.EventToString(a):a for a in pressed}
    if 'LEFTSHIFTKEY' in key_dict.keys() or 'RIGHTSHIFTKEY' in key_dict.keys():
        shift = 1
    else:
        shift = 0
    string = ''
    for char in pressed:
        string += bge.events.EventToCharacter(char, shift)
    return string'''

'''------------------ Non Publically Accessable Functions ------------------'''


def _get_key_buttons():
    '''OUTPUTS:
     - ['AKEY', 'BKEY', 'CKEY']'''
    keys = bge.logic.keyboard.active_events.keys()
    return [bge.events.EventToString(a) for a in keys]

def _get_mouse_buttons():
    '''OUTPUTS:
     - ['LEFTMOUSE', 'RIGHTMOUSE']'''
    active_buttons = list(bge.logic.mouse.active_events.keys())
    all_events = [bge.events.EventToString(a) for a in active_buttons]
    return [e for e in all_events if e not in ['MOUSEX', 'MOUSEY']]
    
def _get_joystick_buttons():
    '''OUTPUTS:
     - ['JOY1BUTTON1', 'JOY2HATUP']'''
    events = []
    i = 0
    for joystick in bge.logic.joysticks:      
        if joystick == None:
            continue
        events += ['JOY{}BUTTON{}'.format(i, b) for b in joystick.activeButtons]
        
        #Hats:  Note that this part may be incorrect? Need to test!!!
        for hat in joystick.hatValues:
            events += ['JOY{}HAT{}'.format(i,HAT_DICT[hat])]
        i += 1
    return events

def _get_button_input():
    '''Returns a dictionary of [action] = percent'''
    active_buttons = []
    active_buttons += _get_key_buttons()
    active_buttons += _get_mouse_buttons()
    active_buttons += _get_joystick_buttons()
    
    status = {}
    for action in _key_map:
        status[action] = 0
        keys = _key_map[action].keys()
        for key in keys:
            if key in active_buttons:
                status[action] += max(-1, min(1, _key_map[action][key]))
    return status





def _get_axis_input():
    '''Returns a dictionary of [action] = percent'''
    all_axis = {}
    all_axis.update(_get_mouse_axis())
    all_axis.update(_get_joy_axis())
    
    status = {}
    for action in _key_map:
        status[action] = 0
        keys = _key_map[action].keys()
        for key in keys:
            if key in all_axis:
                status[action] += max(-1, min(1, all_axis[key]*_key_map[action][key]))
    return status
    

def _get_mouse_axis():
    '''Returns [axis] = percent'''
    pos = bge.logic.mouse.position
    x = pos[0]-0.5
    y = pos[1]-0.5
    
    #if x <= -0.5 or x >= 0.5 or y <= -0.5 or y >= 0.5: #Outside the window, in theory...
    #    x = 0
    #    y = 0
    if center_mouse == True and x != 0 and y != 0:
        bge.logic.mouse.position = (0.5,0.5)
    return {'MOUSEX':x*2,'MOUSEY':y*2}
    
def _get_joy_axis():
    '''Returns [axis] = percent'''
    axis = {}
    j = 0
    for joy in bge.logic.joysticks:
        if joy == None:
            break
        else:
            a = 0
            for value in joy.axisValues:
                axis['JOY{}AXIS{}'.format(j,a)] = value
                a += 1
        j += 1
    return axis


